/*

    Program: ppm2smap - convert ppm images to the Deskplayer's smap format

    GPL (C) 1999/2000  A. Schiffler

*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

/* Text buffer sizes */

#define LINE_LEN	1024

/* Maximum smap size (in bytes) */

#define MAX_SMAP_SIZE	(64*1024)

/* Configurable variables */

static char INPUTFILENAME[LINE_LEN];
static char OUTPUTFILENAME[LINE_LEN];

/* Maximum size of DP palette table */

#define PALETTESIZE 255

/* Setup defaults for configurable variables */

void
set_defaults ()
{
  strcpy (INPUTFILENAME, "");
  strcpy (OUTPUTFILENAME, "");
}

void
SwapBytes (unsigned char *a, unsigned char *b)
{
  unsigned char dummy;
  dummy = *a;
  *a = *b;
  *b = dummy;
}

/* Main routine */

int
main (int argc, char *argv[])
{
  int i, j, x, y;
  int filenamecount = 0;
  FILE *infile;
  FILE *outfile;
  int xsize, ysize;
  unsigned short smapxsize, smapysize;
  int numcolors;
  char header[LINE_LEN];
  char aline[LINE_LEN];
  int numpixels, numbytes;
  unsigned char *image;		/* raw RGB image */
  unsigned char *dpimage;	/* the DP palettized image - 0 is marker value */
  unsigned char *dpimagemir;	/* rotated image for smap conversion */
  unsigned char *curimageR, *curimageG, *curimageB;
  unsigned char *curdpimage;
  unsigned char *curimageR2, *curimageG2, *curimageB2;
  unsigned char *curdpimage2, *curdpimagemir;
  unsigned char *curpaletteR, *curpaletteG, *curpaletteB;
  unsigned char *palette;
  unsigned char numpalette;

  unsigned char palettebits;
  unsigned char *smapimage, *cursmapimage;
  unsigned int smapsize, smaptotalsize;
  unsigned char repeatsize, repeatcount;
  unsigned char smapheader[4 + 2 * 2 + 1];

  int bitpos;
  unsigned char accumulator;

  /* Print message */
  fprintf (stderr,"ppm2smap - A. Schiffler 1999\n\n");

  /* Initialize variables */
  set_defaults ();

  /* Parse command line */
  for (i = 1; i < argc; i++)
    {
      if (argv[i][0] != '-')
	{
	  /* Set filenames */
	  if (filenamecount == 0)
	    {
	      strcpy (INPUTFILENAME, argv[i]);
	      filenamecount++;
	    }
	  else if (filenamecount == 1)
	    {
	      strcpy (OUTPUTFILENAME, argv[i]);
	      filenamecount++;
	    }
	  else
	    {
	      break;
	    }
	}
      else
	{
	  /* Command line help */
	  printf ("Usage: ppm2smap [Input Filename] [Output Filename]\n");
	  exit (0);
	}
    }

  /* Open file for reading or set stdin */
  if (strlen (INPUTFILENAME) == 0)
    {
      infile = stdin;
      strcpy (INPUTFILENAME, "<stdin>");
    }
  else
    {
      if ((infile = fopen (INPUTFILENAME, "r")) == NULL)
	{
	  fprintf (stderr, "ppm2smap: Cannot open file %s for reading.\n",
		   INPUTFILENAME);
	  exit (-1);
	}
    }

  /* Open file for writing or set stdout */
  if (strlen (OUTPUTFILENAME) == 0)
    {
      outfile = stdout;
      strcpy (OUTPUTFILENAME, "<stdout>");
    }
  else
    {
      if ((outfile = fopen (OUTPUTFILENAME, "w")) == NULL)
	{
	  fprintf (stderr, "ppm2smap: Cannot open file %s for writing.\n",
		   OUTPUTFILENAME);
	  exit (-1);
	}
    }

  /* Read and process file */

  /* Read header */
  fgets (aline, LINE_LEN, infile);
  sscanf (aline, "%s", header);

  /* Dump comment lines */
  fgets (aline, LINE_LEN, infile);
  while (aline[0] == '#')
    {
      fgets (aline, LINE_LEN, infile);
    }
  /* Convert size from line */
  sscanf (aline, "%i %i", &xsize, &ysize);
  /* Convert depth from line */
  fgets (aline, LINE_LEN, infile);
  sscanf (aline, "%i", &numcolors);

  /* Progress */
  fprintf (stderr, "%s: %s  %ix%i  %i colors\n", INPUTFILENAME, header, xsize,
	   ysize, numcolors);

  /* Alloc and read the image */
  numpixels = xsize * ysize;
  numbytes = numpixels * 3;
  image = (unsigned char *) calloc (1, numbytes);
  if ((int) fread (image, numbytes, 1, infile) < 1)
    {
      fprintf (stderr, "ppm2smap: cannot read %i bytes from file %s.\n",
	       numbytes, INPUTFILENAME);
      exit (-1);
    }

  /* Alloc space for palette */
  palette = (unsigned char *) calloc (1, 3 * (int) PALETTESIZE);
  numpalette = 0;

  /* Alloc space for palettized image */
  dpimage = (unsigned char *) calloc (1, numpixels);

  /* Convert RGB image to palettized image and update palette as we scan */
  curimageR = image;
  curimageG = image;
  curimageG += 1;
  curimageB = image;
  curimageB += 2;
  curpaletteR = palette;
  curpaletteG = palette;
  curpaletteG += 1;
  curpaletteB = palette;
  curpaletteB += 2;
  curdpimage = dpimage;
  for (i = 0; i < numpixels; i++)
    {
      /* Check for pixel that has not been marked */
      if (*curdpimage == 0)
	{
	  /* Store this color in palette */
	  *curpaletteR = *curimageR;
	  *curpaletteG = *curimageG;
	  *curpaletteB = *curimageB;
	  /* Mark this and all other pixels for this entry */
	  curimageR2 = curimageR;
	  curimageG2 = curimageG;
	  curimageB2 = curimageB;
	  curdpimage2 = curdpimage;
	  /* Update palette counter */
	  numpalette++;
	  /* Inner loop for remaining pixels */
	  for (j = i; j < numpixels; j++)
	    {
	      if ((*curdpimage2 == 0) &&
		  ((unsigned char) (*curimageR2) ==
		   (unsigned char) (*curpaletteR))
		  && ((unsigned char) (*curimageG2) ==
		      (unsigned char) (*curpaletteG))
		  && ((unsigned char) (*curimageB2) ==
		      (unsigned char) (*curpaletteB)))
		{
		  /* Set current palette number in palette image */
		  *curdpimage2 = numpalette;
		}
	      /* Advance positions */
	      curimageR2 += 3;
	      curimageG2 += 3;
	      curimageB2 += 3;
	      curdpimage2++;
	    }
	  /* Update palette pointers */
	  curpaletteR++;
	  curpaletteG++;
	  curpaletteB++;
	}
      /* Advance positions */
      curimageR += 3;
      curimageG += 3;
      curimageB += 3;
      curdpimage++;
      /* Check palette size */
      if (numpalette == 255)
	{
	  /* Exit loop - ignore remaining pixels */
	  i = numpixels;
	}
    }

#ifdef DEBUG
  fprintf (stderr, "\nFound %i different colors.\n", numpalette);
#endif

  /* Prepare smap conversion buffer */
  smapimage = malloc (MAX_SMAP_SIZE + 3);
  dpimagemir = malloc (numpixels);

  /* Calculate number of bits in palette */
  palettebits = 7;
  repeatsize = 255;
  while ((numpalette >> palettebits) == 0)
    {
      palettebits--;
      repeatsize = (repeatsize >> 1);
    }
  palettebits++;
#ifdef DEBUG
  fprintf (stderr, "\nNumber of colors in Palette = %i\n", numpalette);
  fprintf (stderr, "Number of bits in Palette = %i\n", palettebits);
  fprintf (stderr, "Maximum repeatsize = %i\n\n", repeatsize);
#endif

  /* Mirror image */
  curdpimagemir = dpimagemir;
  for (x = 0; x < xsize; x++)
    {
      curdpimage = dpimage;
      curdpimage += x;
      for (y = 0; y < ysize; y++)
	{
	  *curdpimagemir = *curdpimage;
	  curdpimagemir++;
	  curdpimage += xsize;
	}
    }

  /* Convert to smap encoding */
  accumulator = 0;
  bitpos = 8;
  /* Setup source and target pointers */
  curdpimagemir = dpimagemir;	/* Source */
  cursmapimage = smapimage;	/* Target */
  smapsize = 0;			/* Targetsize */
  /* Repeat encoding is only required for >3 colors in palette */
  if (numpalette > 3)
    {
      /* Loop over all pixels */
      for (i = 0; i < numpixels; i++)
	{
	  /* Convert all pixels with repeat-encoding */

	  /* Check if we have repeated pixels */
	  curdpimage = curdpimagemir;	/* Reference pixel */
	  j = i;		/* Next pixel position */
	  curdpimage2 = curdpimagemir;	/* Next pixel */
	  repeatcount = 0;	/* Number of repetitions */
	  while ((j < numpixels) && (*curdpimage == *curdpimage2)
		 && (repeatcount < repeatsize))
	    {
	      curdpimage2++;
	      repeatcount++;
	      j++;
	    }

	  /* push *curdpimagemir */
	  if (bitpos > palettebits)
	    {
	      accumulator |= (*curdpimagemir << (bitpos - palettebits));
	      bitpos -= palettebits;
	    }
	  else if (bitpos == palettebits)
	    {
	      accumulator |= *curdpimagemir;
	      *cursmapimage = accumulator;
	      cursmapimage++;
	      smapsize++;
	      accumulator = 0;
	      bitpos = 8;
	    }
	  else
	    {
	      accumulator |= (*curdpimagemir >> (palettebits - bitpos));
	      *cursmapimage = accumulator;
	      cursmapimage++;
	      smapsize++;
	      accumulator =
		((*curdpimagemir << (8 - (palettebits - bitpos))) & 255);
	      bitpos = (8 - (palettebits - bitpos));
	    }

	  /* repeated-pixel-limit reached, send special sequence */
	  if (repeatcount > 3)
	    {

	      /* push 0 */
	      if (bitpos > palettebits)
		{
		  bitpos -= palettebits;
		}
	      else if (bitpos == palettebits)
		{
		  *cursmapimage = accumulator;
		  cursmapimage++;
		  smapsize++;
		  accumulator = 0;
		  bitpos = 8;
		}
	      else
		{
		  *cursmapimage = accumulator;
		  cursmapimage++;
		  smapsize++;
		  accumulator = 0;
		  bitpos = (8 - (palettebits - bitpos));
		}

	      /* push repeatcount */
	      if (bitpos > palettebits)
		{
		  accumulator |= (repeatcount << (bitpos - palettebits));
		  bitpos -= palettebits;
		}
	      else if (bitpos == palettebits)
		{
		  accumulator |= repeatcount;
		  *cursmapimage = accumulator;
		  cursmapimage++;
		  smapsize++;
		  accumulator = 0;
		  bitpos = 8;
		}
	      else
		{
		  accumulator |= (repeatcount >> (palettebits - bitpos));
		  *cursmapimage = accumulator;
		  cursmapimage++;
		  smapsize++;
		  accumulator =
		    ((repeatcount << (8 - (palettebits - bitpos))) & 255);
		  bitpos = (8 - (palettebits - bitpos));
		}
	      /* advance in source by repeatcount pixels */
	      curdpimagemir += repeatcount;
	      i += (repeatcount - 1);
	    }
	  else
	    {
	      /* advance in source by one pixel */
	      curdpimagemir++;
	    }
	  /* Check smapsize */
	  if (smapsize >= MAX_SMAP_SIZE)
	    {
	      fprintf (stderr, "Output image to large\n");
	      exit (-1);
	    }

	}

    }
  else
    {

      /* Convert all pixels without repeat-encoding */
      for (i = 0; i < numpixels; i++)
	{
	  /* push *curdpimagemir */
	  if (bitpos > palettebits)
	    {
	      accumulator |= (*curdpimagemir << (bitpos - palettebits));
	      bitpos -= palettebits;
	    }
	  else if (bitpos == palettebits)
	    {
	      accumulator |= *curdpimagemir;
	      *smapimage = accumulator;
	      smapimage++;
	      smapsize++;
	      accumulator = 0;
	      bitpos = 8;
	    }
	  else
	    {
	      accumulator |= (*curdpimagemir >> (palettebits - bitpos));
	      *smapimage = accumulator;
	      smapimage++;
	      smapsize++;
	      accumulator =
		((*curdpimagemir << (8 - (palettebits - bitpos))) & 255);
	      bitpos = (8 - (palettebits - bitpos));
	    }
	  /* advance in source by one pixel */
	  curdpimagemir++;
	  /* Check smapsize */
	  if (smapsize >= MAX_SMAP_SIZE)
	    {
	      fprintf (stderr, "Output image to large\n");
	      exit (-1);
	    }
	}			/* for i */
    }

  /* Build smap header */
  smaptotalsize = smapsize + 2 * 2 + 1 + 3 * numpalette;
  smapxsize = xsize;
  smapysize = ysize;
  memcpy (&smapheader[0], &smaptotalsize, 4);
  memcpy (&smapheader[4], &smapxsize, 2);
  memcpy (&smapheader[6], &smapysize, 2);
  memcpy (&smapheader[8], &numpalette, 1);

  SwapBytes (&smapheader[0], &smapheader[3]);
  SwapBytes (&smapheader[1], &smapheader[2]);
  SwapBytes (&smapheader[4], &smapheader[5]);
  SwapBytes (&smapheader[6], &smapheader[7]);

  /* Write out image in DP format */
  fwrite (smapheader, (4 + 2 * 2 + 1), 1, outfile);
  /* Palette */
  curpaletteR = palette;
  curpaletteG = palette + 1;
  curpaletteB = palette + 2;
  for (i = 0; i < numpalette; i++)
    {
      fwrite (curpaletteR, 1, 1, outfile);
      fwrite (curpaletteG, 1, 1, outfile);
      fwrite (curpaletteB, 1, 1, outfile);
      curpaletteR++;
      curpaletteG++;
      curpaletteB++;
    }
  fwrite (smapimage, smapsize, 1, outfile);
  fflush (outfile);

  /* Clean up from conversion */
  free (smapimage);
  free (dpimagemir);

#ifdef DEBUG
  /* Header */
  fprintf (stderr, "Header\n");
  fprintf (stderr, "totalsize   = %i\n", smaptotalsize);
  fprintf (stderr, "smapsize    = %i\n", smapsize);
  fprintf (stderr, "xsize       = %i\n", xsize);
  fprintf (stderr, "ysize       = %i\n", ysize);
  fprintf (stderr, "numpalette  = %i\n", numpalette);
  /* Palette */
  curpaletteR = palette;
  curpaletteG = palette + 1;
  curpaletteB = palette + 2;
  for (i = 0; i < numpalette; i++)
    {
      fprintf (stderr, "Palette %i = %i %i %i\n", i, *curpaletteR,
	       *curpaletteG, *curpaletteB);
      curpaletteR++;
      curpaletteG++;
      curpaletteB++;
    }
#endif

  /* Free memory */
  free (image);
  free (dpimage);

  /* Close files */
  fclose (infile);
  fclose (outfile);

  exit (0);
}
